#! /usr/bin/perl

my %DNA_code = (
'GCT' => 'A', 'GCC' => 'A', 'GCA' => 'A', 'GCG' => 'A', 'TTA' => 'L',
'TTG' => 'L', 'CTT' => 'L', 'CTC' => 'L', 'CTA' => 'L', 'CTG' => 'L',
'CGT' => 'R', 'CGC' => 'R', 'CGA' => 'R', 'CGG' => 'R', 'AGA' => 'R',
'AGG' => 'R', 'AAA' => 'K', 'AAG' => 'K', 'AAT' => 'N', 'AAC' => 'N',
'ATG' => 'M', 'GAT' => 'D', 'GAC' => 'D', 'TTT' => 'F', 'TTC' => 'F',
'TGT' => 'C', 'TGC' => 'C', 'CCT' => 'P', 'CCC' => 'P', 'CCA' => 'P',
'CCG' => 'P', 'CAA' => 'Q', 'CAG' => 'Q', 'TCT' => 'S', 'TCC' => 'S',
'TCA' => 'S', 'TCG' => 'S', 'AGT' => 'S', 'AGC' => 'S', 'GAA' => 'E',
'GAG' => 'E', 'ACT' => 'T', 'ACC' => 'T', 'ACA' => 'T', 'ACG' => 'T',
'GGT' => 'G', 'GGC' => 'G', 'GGA' => 'G', 'GGG' => 'G', 'TGG' => 'W',
'CAT' => 'H', 'CAC' => 'H', 'TAT' => 'Y', 'TAC' => 'Y', 'ATT' => 'I',
'ATC' => 'I', 'ATA' => 'I', 'GTT' => 'V', 'GTC' => 'V', 'GTA' => 'V',
'GTG' => 'V', 'TAA' => '*', 'TGA' => '*', 'TAG' => '*');

my %SNP_code = ('NS' => 0, 'STOP' => 0, 'S' => 1, 'NC' => 2, 'INDEL' => 3, 'indel' => 3, 'INS' => 3, 'DEL' => 3, 'FS' => 3);
my @bases = ("A", "C", "G", "T");
my $maxAbsent = 0;  ## maximum allowed missings for each site

$genome = shift;
$gene = shift;
$variation = shift;

### read in reference genome ###
local %seq;
local @names;
open(REF, "<$genome") or die;
while (<REF>) {
	chomp $_;
	if($_ =~ /^>(\S+)/) {
		$names[@names] = $1;
	} else {
		$seq{$names[@names-1]} .= $_;
	}
}
close REF;
foreach $name (@names) {
	$seq{$name} =~ tr/a-z/A-Z/;
}

local @matrix;
local @snps;
for(local $i=0; $i<length($seq{$names[0]}); ++$i) {
	@{$matrix[$i]} = (0, 0, 0, 1);
	@{$snps[$i]} = (0, 0, 0, 0);
}

### read in SNPs/indels ####
my %mut_table;
foreach $ori (@bases) {
	foreach $dev (@bases) {
		if($ori ne $dev) {
			${$mut_table{$ori}}{$dev} = 5;
		}
	}
}

my $mut_rate = 0;
open(VAR, "<$variation") or die;
while (<VAR>) {
	chomp $_;
	@part = split /\s/, $_;
	$part[1] =~ tr/a-z/A-Z/; $part[2] =~ tr/a-z/A-Z/;
	${$snps[$part[0]-1]}[$SNP_code{$part[3]}] ++;
	if($SNP_code{$part[3]} < 3) {
		${$mut_table{$part[1]}}{$part[2]} ++;
		${$mut_table{$part[2]}}{$part[1]} ++;
		$mut_rate ++;
	}
}

$mut_rate = $mut_rate/length($seq{$names[0]});
my @mut_dist = ((1-$mut_rate)*(1-$mut_rate), 2*(1-$mut_rate)*$mut_rate, $mut_rate*$mut_rate);
my %code_rate;
foreach $c1 (keys %DNA_code) {
	local @crate;
	foreach $c2 (keys %DNA_code) {
		local $aa = 0; 
		if($DNA_code{$c1} ne $DNA_code{$c2}) {
			if($DNA_code{$c1} eq "*" || $DNA_code{$c2} eq "*") {$aa = 2;}
			else {$aa = 1;}
		}
		local $n = -1; local $weight = 1; local @distance;
		for(local $i=0; $i<3; ++$i) {
			$distance[$i] = ${$mut_table{substr($c1, $i, 1)}}{substr($c2, $i, 1)};
			if($distance[$i] > 0) {$weight *= $distance[$i]; $n ++;}
		}
		for(local $i=0; $i<3; ++$i) {
			if($distance[$i] > 0) {$crate[$i][$n][$aa] += $weight;}
		}
	}
	for(local $i=0; $i<3; ++$i) {
		for(local $j=0; $j<3; ++$j) {
			local $total = $crate[$i][$j][0]+$crate[$i][$j][1]+$crate[$i][$j][2];
			$crate[$i][$j][0] = $crate[$i][$j][0]/$total;
			$crate[$i][$j][1] = $crate[$i][$j][1]/$total;
			$crate[$i][$j][2] = $crate[$i][$j][2]/$total;
		}
		@{$code_rate{"$c1 $i"}} =  ($crate[$i][0][0]*$mut_dist[0]+$crate[$i][1][0]*$mut_dist[1]+$crate[$i][2][0]*$mut_dist[2],
									$crate[$i][0][1]*$mut_dist[0]+$crate[$i][1][1]*$mut_dist[1]+$crate[$i][2][1]*$mut_dist[2],
									$crate[$i][0][2]*$mut_dist[0]+$crate[$i][1][2]*$mut_dist[1]+$crate[$i][2][2]*$mut_dist[2]);
	}
}

open (GG, "<$gene") or die;
while (<GG>) {
	chomp $_;
	local @part = split /\s+/, $_;
	if($part[2] eq "CDS") {
		@{$cds[@cds]} = ($part[0], $part[1], $part[3], $part[4], $part[6]);
	}
}
close GG;

foreach $g (@cds) {
	while ( ($$g[3]-$$g[2]+1)%3 != 0) {
		if($$g[4] eq "+") {$$g[3] --;}
		else {$$g[2] ++;}
	}
	if ($$g[4] eq "+") {
		for(local $i=0; $i < $$g[3]-$$g[2]+1; $i+=3) {
			local $site = $$g[2] + $i;
			local %codons; local $variant = 0;
			if( $snps[$site-1][0]+$snps[$site-1][1]+$snps[$site-1][2]
				+$snps[$site][0]+$snps[$site][1]+$snps[$site][2]
				+$snps[$site+1][0]+$snps[$site+1][1]+$snps[$site+1][2] == 0 ) {
				$variant = 0;
			} else {
				$variant = 1;
			}
			local $ref_codon = substr($seq{$names[0]}, $site-1, 3);
			for(local $j=0; $j<@names; ++$j) {
				local $codon = substr($seq{$names[$j]}, $site-1, 3);
				if($codon =~ /\-\-/) {
					$codons{$codon} ++;
				} elsif ($variant == 1) {
					$codons{$codon} ++;
				} else {
					$codons{$ref_codon} ++;
				}
			}

			local @change = determineNS(\%codons);
			$matrix[$site-1][0] += $change[0]; $matrix[$site-1][1] += $change[1]; $matrix[$site-1][2] += $change[2]; $matrix[$site-1][3] = 0;
			$matrix[$site][0] += $change[3]; $matrix[$site][1] += $change[4]; $matrix[$site][2] += $change[5]; $matrix[$site][3] = 0;
			$matrix[$site+1][0] += $change[6]; $matrix[$site+1][1] += $change[7]; $matrix[$site+1][2] += $change[8]; $matrix[$site+1][3] = 0;
		}
	} else {
		for(local $i=0; $i < $$g[3]-$$g[2]+1; $i+=3) {
			local $site = $$g[3] - $i;
			local %codons; local $variant = 0;
			if( $snps[$site-1][0]+$snps[$site-1][1]+$snps[$site-1][2]
				+$snps[$site-2][0]+$snps[$site-2][1]+$snps[$site-2][2]
				+$snps[$site-3][0]+$snps[$site-3][1]+$snps[$site-3][2] == 0 ) {
				$variant = 0;
			} else {
				$variant = 1;
			}
			local $ref_codon = substr($seq{$names[0]}, $site-3, 3); $ref_codon =~ tr/ATCG/TAGC/; $ref_codon = reverse $ref_codon;
			for(local $j=0; $j<@names; ++$j) {
				local $codon = substr($seq{$names[$j]}, $site-3, 3);
				if($codon =~ /\-\-/) {
					$codons{$codon} ++;
				} elsif ($variant == 1) {
					$codon =~ tr/ATCG/TAGC/; $codon = reverse $codon;
					$codons{$codon} ++;
				} else {
					$codons{$ref_codon} ++;
				}
			}

			local @change = determineNS(\%codons);
			# print STDERR join(" ", keys(%codons))."  $site  @change\n"; exit;
			$matrix[$site-1][0] += $change[0]; $matrix[$site-1][1] += $change[1]; $matrix[$site-1][2] += $change[2]; $matrix[$site-1][3] = 0;
			$matrix[$site-2][0] += $change[3]; $matrix[$site-2][1] += $change[4]; $matrix[$site-2][2] += $change[5]; $matrix[$site-2][3] = 0;
			$matrix[$site-3][0] += $change[6]; $matrix[$site-3][1] += $change[7]; $matrix[$site-3][2] += $change[8]; $matrix[$site-3][3] = 0;
		}
	}
}

for(local $i=0; $i<@matrix; ++$i) {
	if($matrix[$i][3] == 1) {
		local %type;
		for(local $j=0; $j<@names; ++$j) {
			$type{substr($seq{$names[$j]}, $i, 1)} ++;
		}
		local $absent = 0; local $types = 0;
		foreach $t (keys %type) {
			if($t =~ /[A-Z]/) {
				$types ++; 
			} else {
				$absent += $type{$t};
			}
		}
		if($absent > $maxAbsent) {$matrix[$i][3] = 0;}
		elsif($types > 1 && $variation eq "") {
			$snps[$i][2] = 1;
		}
	}
	if($matrix[$i][0] + $matrix[$i][1] + $matrix[$i][2] > 0) {
		local $total = $matrix[$i][0] + $matrix[$i][1] + $matrix[$i][2];
		$matrix[$i][0] = $matrix[$i][0]/$total; $dn += $matrix[$i][0];
		$matrix[$i][1] = $matrix[$i][1]/$total; $ds += $matrix[$i][1];
		$matrix[$i][2] = $matrix[$i][2]/$total; $dstop += $matrix[$i][2];
	} else {
		$dnc += $matrix[$i][3];
	}

	@test = ($matrix[$i][0]+$matrix[$i][2], $matrix[$i][1], $matrix[$i][3], $snps[$i][0], $snps[$i][1], $snps[$i][2], $snps[$i][3]);
	if($test[0] == 0 && $test[3] > 0) {print STDERR "WARNING: NS snp in ZERO possiblility ".($i+1)."\n"; $test[3] = 0;}
	if($test[1] == 0 && $test[4] > 0) {print STDERR "WARNING: S snp in ZERO possiblility ".($i+1)."\n"; $test[4] = 0;}
	if($test[2] == 0 && $test[5] > 0) {print STDERR "WARNING: NC snp in ZERO possiblility ".($i+1)."\n"; $test[5] = 0;}
	print "@test\n";
}
print STDERR "NS site: ".($dn+$dstop)."\tS site: $ds\tSTOP site: $dstop\tNC site: $dnc\n";

sub determineNS {
	local $codon = $_[0];
	local @codons; local $absent = 0;
	foreach $c (keys(%$codon)) {
		if($c =~ /\-\-/) {
			$absent += $$codon{$c};
		} elsif(!($c =~/[^ATCG]/)) {
			$codons[@codons] = $c;
		}
	}
	local @result = (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
	if($absent > $maxAbsent) { return @result; }
	foreach $c (@codons) {
		for(local $cpos=0; $cpos<3; ++$cpos) {
			$result[0 + $cpos*3] += ${$code_rate{"$c $cpos"}}[1]/scalar(@codons);
			$result[1 + $cpos*3] += ${$code_rate{"$c $cpos"}}[0]/scalar(@codons);
			$result[2 + $cpos*3] += ${$code_rate{"$c $cpos"}}[2]/scalar(@codons);
		}
	}
	return @result;
}
